"""IIIF Authentication for the external interaction pattern.

See <http://iiif.io/api/auth/#external-interaction-pattern>.
Trivial code to immediately close the login window and
set the account cookie.
"""

from .auth_flask import IIIFAuthFlask


class IIIFAuthExternal(IIIFAuthFlask):
    """IIIF Authentication class using a simple external."""

    def __init__(self, cookie_prefix=None):
        """Initialize IIIFAuthBasic object."""
        super(IIIFAuthExternal, self).__init__(cookie_prefix=cookie_prefix)
        self.auth_pattern = 'external'
        self.label = 'External Authentication Required'
        self.failureHeader = 'Restricted Material'
        self.failureDescription = 'This material is not viewable without prior agreement'

    def login_service_description(self):
        """External login service description.

        The login service description _MUST_ include the token service
        description. For the external interaction pattern we have
        a dummy URI for the access cookie service.
        """
        desc = super(IIIFAuthExternal, self).login_service_description()
        return desc

    def login_handler(self, config=None, prefix=None, **args):
        """Null login handler for external.

        We provide no support for getting the access cookie,
        this must be done by the use via some out-of-band means.
        """
        return True

    # No logout for external
    logout_handler = None



"""IIIF Authentication using the login interaction pattern via Google.

See <http://iiif.io/api/auth/#login-interaction-pattern>.
This code is specifc to the Flask webapp framework.
"""

import json
try:
    # python3
    from urllib.request import urlopen
    from urllib.request import Request
    from urllib.parse import urlencode
except ImportError:
    # fall back to python2
    from urllib2 import urlopen
    from urllib2 import Request
    from urllib import urlencode

from .auth_flask import IIIFAuthFlask


class IIIFAuthGoogle(IIIFAuthFlask):
    """IIIF Authentication Class using Google Auth."""

    def __init__(self, client_secret_file='client_secret.json',
                 cookie_prefix=None):
        """Initialize IIIFAuthGoogle object.

        Sets a number of constants and attempts to load client client_secret
        data from client_secret_file.
        """
        super(IIIFAuthGoogle, self).__init__(cookie_prefix=cookie_prefix)
        #
        self.auth_pattern = 'login'
        self.auth_type = 'Google auth'
        try:
            # Assign defaults so code/tests will have some data even if load
            # fails
            self.google_api_scope = (  # pep8 long string...
                'https://www.googleapis.com/auth/userinfo.profile '
                'https://www.googleapis.com/auth/userinfo.email')
            self.google_oauth2_url = 'https://accounts.google.com/o/oauth2/'
            self.google_api_url = 'https://www.googleapis.com/oauth2/v1/'
            self.google_api_client_id = 'oops_missing_client_id'
            self.google_api_client_secret = 'oops_missing_client_secret'
            gcd = json.loads(open(client_secret_file).read())
            self.google_api_client_id = gcd['web']['client_id']
            self.google_api_client_secret = gcd['web']['client_secret']
        except Exception as e:
            self.logger.error("Failed to load Google auth from %s: %s" % (
                client_secret_file, str(e)))

    def login_handler(self, config=None, prefix=None, **args):
        """OAuth starts here, redirect user to Google."""
        params = {
            'response_type': 'code',
            'client_id': self.google_api_client_id,
            'redirect_uri': self.scheme_host_port_prefix(
                'http', config.host, config.port, prefix) + '/home',
            'scope': self.google_api_scope,
            'state': self.request_args_get('next', default=''),
        }
        url = self.google_oauth2_url + 'auth?' + urlencode(params)
        return self.login_handler_redirect(url)

    def home_handler(self, config=None, prefix=None, **args):
        """Handler for /home redirect path after Google auth.

        OAuth ends up back here from Google. Set the account cookie
        and close window to trigger next step.
        """
        gresponse = self.google_get_token(config, prefix)
        gdata = self.google_get_data(config, gresponse)
        email = gdata.get('email', 'NO_EMAIL')
        name = gdata.get('name', 'NO_NAME')
        # Make and store cookie from identity, set and close window
        cookie = self.access_cookie(name + ' ' + email)
        return self.set_cookie_close_window_response(cookie)

    ######################################################################
    # Code to get data from Google API
    #

    def google_get_token(self, config, prefix):
        """Make request to Google API to get token."""
        params = {
            'code': self.request_args_get(
                'code',
                default=''),
            'client_id': self.google_api_client_id,
            'client_secret': self.google_api_client_secret,
            'redirect_uri': self.scheme_host_port_prefix(
                'http', config.host, config.port, prefix) + '/home',
            'grant_type': 'authorization_code',
        }
        payload = urlencode(params).encode('utf-8')
        url = self.google_oauth2_url + 'token'
        req = Request(url, payload)
        json_str = urlopen(req).read()
        return json.loads(json_str.decode('utf-8'))

    def google_get_data(self, config, response):
        """Make request to Google API to get profile data for the user."""
        params = {
            'access_token': response['access_token'],
        }
        payload = urlencode(params)
        url = self.google_api_url + 'userinfo?' + payload
        req = Request(url)
        json_str = urlopen(req).read()
        return json.loads(json_str.decode('utf-8'))



"""IIIF Image Information Response.

Model for IIIF Image API 'Image Information Response'.
Default version is 2.1 but also supports 3.0, 2.0, 1.1 and 1.0.

Philisophy is to migrate this code forward with new versions
of the specification but to keep support for all published
versions.
"""

import sys
import json
import re


def _parse_string_or_array(json_data):
    # Parse either JSON single string or array into array.
    if (not isinstance(json_data, list)):
        json_data = [json_data]
    return json_data


def _parse_int_array(json_data):
    # Force simple array of interger values.
    return [int(x) for x in json_data]


def _parse_noop(json_data):
    # Format is already what we want.
    return json_data


def _parse_sizes(json_data):
    # Parse sizes in 2.0 and above.
    #
    # 3.0 spec: "A list of JSON objects with the height and width properties.
    # These sizes specify preferred values to be provided in the w,h syntax of
    # the size request parameter for scaled versions of the full image."
    if (not isinstance(json_data, list)):
        raise IIIFInfoError("The sizes property have a list value")
    for obj in json_data:
        if (not isinstance(obj, dict) or "width" not in obj or "height" not in obj):
            raise IIIFInfoError("Every entry in the sizes property list must have width and height")
    return json_data


def _parse_tile(json_data):
    # Parse data for a single tile specification.
    tile = {}
    tile['width'] = int(json_data['width'])
    if ('height' in json_data):
        tile['height'] = int(json_data['height'])
    tile['scaleFactors'] = _parse_int_array(json_data['scaleFactors'])
    return tile


def _parse_tiles(json_data):
    # Parse tiles array of tile specifications.
    #
    # Expect common case in 2.0 and above to map to 1.1 idea of
    # tile_width, tile_height and scale_factors. This is the case when
    # len()==1.
    tiles = []
    if (len(json_data) == 0):
        raise IIIFInfoError("Empty tiles array property not allowed.")
    for tile_spec in json_data:
        tiles.append(_parse_tile(tile_spec))
    return tiles


def _parse_service(json_data):
    return json_data


def _parse_profile_3_x(json_data):
    # 3.0 spec: "A string indicating the highest compliance level which is
    # fully supported by the service. The value must be one of "level0",
    # "level1", or "level2"."
    if (json_data not in ("level0", "level1", "level2")):
        raise IIIFInfoError("The value of the profile property must be a level string")
    return json_data


def _parse_profile_2_x(json_data):
    # 2.1 spec: "A list of profiles, indicated by either a URI or an
    # object describing the features supported. The first entry
    # in the list must be a compliance level URI."
    #
    # 2.0 spec: "An array of profiles, indicated by either a URI or
    # an object describing the features supported. The first entry in
    # the array must be a compliance level URI, as defined below."
    if (not isinstance(json_data, list)):
        raise IIIFInfoError("The profile property have a list value")
    if (not re.match(r'''http://iiif.io/api/image/2/level[012]\.json''', json_data[0])):
        raise IIIFInfoError("The first entry in the profile list must be a compliance level URI.")
    return json_data


def _parse_profile_1_1(json_data):
    # 1.1 & 1.0 spec: "URI indicating the compliance level supported.
    # Values as described in Section 8. Compliance Levels"
    if (not re.match(r'''http://library.stanford.edu/iiif/image-api/1.1/compliance.html#level[012]''', json_data)):
        raise IIIFInfoError("The profile property value must be a compliance level URI.")
    return json_data


def _parse_profile_1_0(json_data):
    # 1.1 & 1.0 spec: "URI indicating the compliance level supported.
    # Values as described in Section 8. Compliance Levels"
    if (not re.match(r'''http://library.stanford.edu/iiif/image-api/compliance.html#level[012]''', json_data)):
        raise IIIFInfoError("The profile property value must be a compliance level URI.")
    return json_data


# Configuration information for API versions

CONF = {
    '1.0': {
        'params': [
            'id', 'protocol', 'width', 'height',
            'scale_factors', 'tile_width', 'tile_height',
            'extra_formats', 'extra_qualities', 'profile'
        ],
        'array_params': set([
            'scale_factors', 'extra_formats', 'extra_qualities'
        ]),
        'complex_params': {
            'profile': _parse_profile_1_0,
            'scale_factors': _parse_int_array,
            'extra_formats': _parse_noop,  # array of str
            'extra_qualities': _parse_noop  # array of str
        },
        'compliance_prefix': "http://library.stanford.edu/iiif/image-api/compliance.html#level",
        'compliance_suffix': "",
        'protocol': None,
        'required_params': ['id', 'width', 'height', 'profile'],
        'property_to_json': {
            'id': 'identifier',
            'extra_formats': 'formats',
            'extra_qualities': 'qualities'
        }
    },
    '1.1': {
        'params': [
            'id', 'protocol', 'width', 'height',
            'scale_factors', 'tile_width', 'tile_height',
            'extra_formats', 'extra_qualities', 'profile'
        ],
        'array_params': set([
            'scale_factors', 'extra_formats', 'extra_qualities'
        ]),
        'complex_params': {
            'profile': _parse_profile_1_1,
            'scale_factors': _parse_int_array,
            'extra_formats': _parse_noop,  # array of str
            'extra_qualities': _parse_noop  # array of str
        },
        'api_context': "http://library.stanford.edu/iiif/image-api/1.1/context.json",
        'compliance_prefix': "http://library.stanford.edu/iiif/image-api/1.1/compliance.html#level",
        'compliance_suffix': "",
        'protocol': None,
        'required_params': [
            'id', 'width', 'height', 'profile'
        ],
        'property_to_json': {
            'id': '@id',
            'extra_formats': 'formats',
            'extra_qualities': 'qualities'
        }
    },
    '2.0': {
        'params': [
            'id', 'protocol', 'width', 'height',
            'profile', 'sizes', 'tiles', 'service'
        ],
        # scale_factors isn't in API but used internally
        'array_params': set([
            'sizes', 'tiles', 'service', 'scale_factors', 'extra_formats',
            'extra_qualities', 'supports'
        ]),
        'complex_params': {
            'sizes': _parse_sizes,
            'tiles': _parse_tiles,
            'profile': _parse_profile_2_x,
            'service': _parse_service
        },
        'api_context': "http://iiif.io/api/image/2/context.json",
        'compliance_prefix': "http://iiif.io/api/image/2/level",
        'compliance_suffix': ".json",
        'protocol': "http://iiif.io/api/image",
        'required_params': [
            'id', 'protocol', 'width', 'height', 'profile'
        ],
        'property_to_json': {
            'id': '@id',
            'extra_formats': 'formats',
            'extra_qualities': 'qualities'
        }
    },
    '2.1': {
        'params': [
            'id', 'protocol', 'width', 'height',
            'profile', 'sizes', 'tiles', 'service',
            'attribution', 'logo', 'license'
        ],
        # scale_factors isn't in API but used internally
        'array_params': set([
            'sizes', 'tiles', 'service', 'scale_factors', 'extra_formats',
            'maxArea', 'maxHeight', 'maxWidth', 'extra_qualities', 'supports'
        ]),
        'complex_params': {
            'sizes': _parse_sizes,
            'tiles': _parse_tiles,
            'profile': _parse_profile_2_x,
            'service': _parse_service
        },
        'api_context': "http://iiif.io/api/image/2/context.json",
        'compliance_prefix': "http://iiif.io/api/image/2/level",
        'compliance_suffix': ".json",
        'protocol': "http://iiif.io/api/image",
        'required_params': [
            'id', 'protocol', 'width', 'height', 'profile'
        ],
        'property_to_json': {
            'id': '@id',
            'extra_formats': 'formats',
            'extra_qualities': 'qualities'
        }
    },
    '3.0': {
        'params': [
            'id', 'resource_type', 'protocol',
            'width', 'height', 'profile', 'sizes', 'tiles',
            'extra_formats', 'extra_qualities', 'extra_features',
            'service', 'attribution', 'logo', 'license'
        ],
        # scale_factors isn't in API but used internally
        'array_params': set([
            'sizes', 'tiles', 'service', 'scale_factors',
            'extra_formats', 'extra_qualities', 'extra_features',
            'maxArea', 'maxHeight', 'maxWidth'
        ]),
        'complex_params': {
            'sizes': _parse_sizes,
            'tiles': _parse_tiles,
            'profile': _parse_profile_3_x,
            'service': _parse_service
        },
        'api_context': "http://iiif.io/api/image/3/context.json",
        'compliance_prefix': "level",
        'compliance_suffix': "",
        'protocol': "http://iiif.io/api/image",
        'required_params': [
            'id', 'protocol', 'width', 'height', 'profile'
        ],
        'property_to_json': {
            'resource_type': 'type',
            'extra_formats': 'extraFormats',
            'extra_qualities': 'extraQualities',
            'extra_features': 'extraFeatures'
        },
        'fixed_values': {
            'resource_type': 'ImageService3'
        }
    }
}


class IIIFInfoError(Exception):
    """IIIFInfoError from IIIFInfo."""

    pass


class IIIFInfoContextError(IIIFInfoError):
    """IIIFInfoContextError for @context issues from IIIFInfo."""

    pass


class IIIFInfo(object):
    """IIIF Image Information Class."""

    def __init__(self, api_version='2.1', profile=None, level=1, conf=None,
                 server_and_prefix='',
                 identifier=None, width=None, height=None, tiles=None,
                 sizes=None, service=None, id=None,
                 # legacy params from 1.1
                 scale_factors=None, tile_width=None, tile_height=None,
                 # 1.1, 2.0, 2.1
                 formats=None, qualities=None,
                 # 2.0 and 2.1 only
                 supports=None,
                 # 2.1 only
                 attribution=None, logo=None, license=None,
                 # 3.0 only
                 extra_formats=None, extra_qualities=None,
                 extra_features=None,
                 ):
        """Initialize an IIIFInfo object.

        Parameters include:
        api_version -- defaults to 2.0 but may be set to 1.1
        profile -- usually not set but overrides handling via level
            (2.0 complex profile not supported except by explicitly
            passing in an array that matches JSON required)
        level -- default 1, generates compliance level for simple profile
        """
        # API version (used in level settings)
        if (api_version not in CONF):
            raise IIIFInfoError(
                "Unknown IIIF Image API version '%s', versions supported are ('%s')" %
                (api_version, sorted(CONF.keys())))
        self.api_version = api_version
        self.set_version_info()
        if (profile is not None):
            # Explicit profile setting overrides any level set
            # or default level
            self.profile = profile
        else:
            self.level = level
        # explicit settings
        self.server_and_prefix = server_and_prefix
        self.identifier = identifier
        self.width = width
        self.height = height
        self.tiles = tiles
        self.sizes = sizes
        self.service = service
        # legacy params for 1.0 and 1.1
        if (scale_factors is not None):
            self.scale_factors = scale_factors
        if (tile_width is not None):
            self.tile_width = tile_width
        if (tile_height is not None):
            self.tile_height = tile_height
        # 1.1+
        self.extra_formats = formats
        self.extra_qualities = qualities
        # 2.0+ only
        self.supports = supports
        # 2.1+ only
        self.attribution = attribution
        self.logo = logo
        self.license = license
        # 3.0 only
        if extra_formats:
            self.extra_formats = extra_formats
        if extra_qualities:
            self.extra_qualities = extra_qualities
        if extra_features:
            self.extra_features = extra_features
        # defaults from conf dict if provided
        if (conf):
            for option in conf:
                setattr(self, option, conf[option])
        if (id is not None):
            self.id = id

    @property
    def context(self):
        """Image API JSON-LD @context.

        Context taken from self.contexts, will return the last
        one if there are multiple, None if not set.
        """
        try:
            return self.contexts[-1]
        except:
            return None

    def add_context(self, context):
        """Add context to the set of contexts if not already present."""
        if context not in self.contexts:
            self.contexts.insert(0, context)

    @property
    def id(self):
        """id property based on server_and_prefix and identifier."""
        id = ''
        if (self.server_and_prefix is not None and
                self.server_and_prefix != ''):
            id += self.server_and_prefix + '/'
        if (self.identifier is not None):
            id += self.identifier
        return id

    @id.setter
    def id(self, value):
        """Split into server_and_prefix and identifier."""
        i = value.rfind('/')
        if (i > 0):
            self.server_and_prefix = value[:i]
            self.identifier = value[(i + 1):]
        elif (i == 0):
            self.server_and_prefix = ''
            self.identifier = value[(i + 1):]
        else:
            self.server_and_prefix = ''
            self.identifier = value

    def set_version_info(self, api_version=None):
        """Set version and load configuration for given api_version.

        Will use current value of self.api_version if a version number
        is not specified in the call. Will raise IIIFInfoError if an
        unknown API version is supplied.

        Sets a number of configuration properties from the content
        of CONF[api_version] which then control much of the rest of
        the behavior of this object.
        """
        if (api_version is None):
            api_version = self.api_version
        if (api_version not in CONF):
            raise IIIFInfoError("Unknown API version %s" % (api_version))
        # Load configuration for version
        self.params = CONF[api_version]['params']
        self.array_params = CONF[api_version]['array_params']
        self.complex_params = CONF[api_version]['complex_params']
        for a in ('api_context', 'compliance_prefix', 'compliance_suffix',
                  'protocol', 'required_params', 'property_to_json',
                  'fixed_values'):
            if (a in CONF[api_version]):
                self._setattr(a, CONF[api_version][a])
                if (a == 'api_context'):
                    self.contexts = [CONF[api_version][a]]
        # Set any fixed values
        if hasattr(self, 'fixed_values'):
            for p, v in self.fixed_values.items():
                self._setattr(p, v)

    def json_key(self, property):
        """JSON key for given object property name.

        If no mapping is specified then the JSON key is assumed to be
        the property name.
        """
        try:
            return self.property_to_json[property]
        except (AttributeError, KeyError):
            return property

    @property
    def compliance(self):
        """Compliance profile URI.

        In IIIF Image API v3.x the profile information is a JSON-LD
        aliased string representing the compliance level URI.

        In IIIF Image API v2.x the profile information is an array
        of values and objects, the first of which must be the
        compliance level URI.

        In IIIF Image API v1.x the profile information is
        just a single profile URI.
        """
        if self.api_version < '2.0':
            return self.profile
        elif self.api_version < '3.0':
            return self.profile[0]
        else:
            return self.profile

    @compliance.setter
    def compliance(self, value):
        """Set the compliance profile URI."""
        if (self.api_version < '2.0'):
            self.profile = value
        elif self.api_version < '3.0':
            try:
                self.profile[0] = value
            except AttributeError:
                # handle case where profile not initialized as array
                self.profile = [value]
        else:
            self.profile = value

    @property
    def level(self):
        """Extract level number from compliance profile URI.

        Returns integer level number or raises IIIFInfoError
        """
        m = re.match(
            self.compliance_prefix +
            r'(\d)' +
            self.compliance_suffix +
            r'$',
            self.compliance)
        if (m):
            return int(m.group(1))
        raise IIIFInfoError(
            "Bad compliance profile URI, failed to extract level number")

    @level.setter
    def level(self, value):
        """Build profile URI from level.

        Level should be an integer 0,1,2
        """
        self.compliance = self.compliance_prefix + \
            ("%d" % value) + self.compliance_suffix

    def _single_tile_getter(self, param):
        # Extract param from a single tileset defintion
        if (self.tiles is None or len(self.tiles) == 0):
            return None
        else:
            return self.tiles[0].get(param, None)

    def _single_tile_setter(self, param, value):
        # Set param for a single tileset defintion
        if (self.tiles is None or len(self.tiles) == 0):
            self.tiles = [{}]
        self.tiles[0][param] = value

    @property
    def scale_factors(self):
        """Access to scale_factors in 1.x.

        Also provides the scale factors in 2.0 and greated for
        only the first tile definition.
        """
        return self._single_tile_getter('scaleFactors')

    @scale_factors.setter
    def scale_factors(self, value):
        """Set scale_factors at version 1.1 and below.

        If used at 2.0+ this will create/edit the first entry
        in self.tiles. Will throw and error if it is used with more
        than on entry for self.tiles.
        """
        self._single_tile_setter('scaleFactors', value)

    @property
    def tile_width(self):
        """Access to tile_width in 1.x.

        Also provides the tile_width in 2.0 and greater for
        only the first tile definition.
        """
        return self._single_tile_getter('width')

    @tile_width.setter
    def tile_width(self, value):
        """Set tile_width at version 1.1 and below."""
        self._single_tile_setter('width', value)

    @property
    def tile_height(self):
        """Access to tile_height in 1.x.

        Also provides the tile_height in 2.0 and greaterfor
        only the first tile definition. If width is set but not
        height then return that instead.
        """
        h = self._single_tile_getter('height')
        if (h is None):
            return self._single_tile_getter('width')
        return h

    @tile_height.setter
    def tile_height(self, value):
        """Set tile_height at version 1.1 and below."""
        self._single_tile_setter('height', value)

    def add_service(self, service):
        """Add a service description.

        Handles transition from self.service=None, self.service=dict for a
        single service, and then self.service=[dict,dict,...] for multiple
        """
        if (self.service is None):
            self.service = service
        elif (isinstance(self.service, dict)):
            self.service = [self.service, service]
        else:
            self.service.append(service)

    def _setattr(self, param, value):
        # Setter handling array/scalar properties, and legacy.
        #
        # Also deals with legacy parameters that map to other
        # variables (but ugly!).
        if (param in self.array_params and
                isinstance(value, str)):
            # If we have an array then set directly, make list. Perhaps not
            # pythonic to do a type check for array here but want to avoid
            # accidentally iterating on chars in string etc..
            value = [value]
        setattr(self, param, value)

    def validate(self):
        """Validate this object as Image API data.

        Raise IIIFInfoError with helpful message if not valid.
        """
        errors = []
        for param in self.required_params:
            if (not hasattr(self, param) or getattr(self, param) is None):
                errors.append("missing %s parameter" % (param))
        if (len(errors) > 0):
            raise IIIFInfoError("Bad data for info.json: " + ", ".join(errors))
        return True

    def as_json(self, validate=True):
        """Return JSON serialization.

        Will raise IIIFInfoError if insufficient parameters are present to
        have a valid info.json response (unless validate is False).
        """
        if (validate):
            self.validate()
        json_dict = {}
        if (self.api_version > '1.0'):
            if (len(self.contexts) > 1):
                json_dict['@context'] = self.contexts
            else:
                json_dict['@context'] = self.context
        params_to_write = set(self.params)
        params_to_write.discard('id')
        if (self.identifier):
            if (self.api_version == '1.0'):
                json_dict['identifier'] = self.identifier  # local id
            else:
                json_dict[self.json_key('id')] = self.id  # URI
        params_to_write.discard('profile')
        if (self.compliance):
            if (self.api_version < '2.0'):
                json_dict['profile'] = self.compliance
            elif (self.api_version < '3.0'):
                # FIXME - need to support extra profile features
                json_dict['profile'] = [self.compliance]
                d = {}
                if (self.extra_formats is not None):
                    d['formats'] = self.extra_formats
                if (self.extra_qualities is not None):
                    d['qualities'] = self.extra_qualities
                if (self.supports is not None):
                    d['supports'] = self.extra_features
                if (len(d) > 0):
                    json_dict['profile'].append(d)
                params_to_write.discard('extra_formats')
                params_to_write.discard('extra_qualities')
                params_to_write.discard('extra_features')
            else:
                json_dict['profile'] = self.profile
        for param in params_to_write:
            if (hasattr(self, param) and
                    getattr(self, param) is not None):
                json_dict[self.json_key(param)] = getattr(self, param)
        return(json.dumps(json_dict, sort_keys=True, indent=2))

    def read_property(self, j, param):
        """Read one property param from JSON j."""
        key = self.json_key(param)
        if (key in j):
            value = j[key]
            if (param in self.complex_params):
                # use function ref in self.complex_params to parse
                value = self.complex_params[param](value)
            self._setattr(param, value)

    def read(self, fh, api_version=None):
        """Read info.json from file like object.

        Parameters:
        fh -- file like object supporting fh.read()
        api_version -- IIIF Image API version expected

        If api_version is set then the parsing will assume this API version,
        else the version will be determined from the incoming data. NOTE that
        the value of self.api_version is NOT used in this routine.

        If an api_version is specified and there is a @context specified then
        an IIIFInfoContextError will be raised unless these match. If no known
        @context is present and no api_version set then an IIIFInfoContextError
        will be raised.
        """
        # load and parse JSON
        j = json.load(fh)
        # must work out API version in order to know how to parse JSON
        # extract image API specific @context and API version
        if (api_version == '1.0'):
            # v1.0 did not have a @context so take the version passed in
            self.api_version = api_version
        else:
            try:
                self.contexts = _parse_string_or_array(j['@context'])
            except KeyError:
                # no @context and not 1.0
                if (api_version is None):
                    raise IIIFInfoContextError("No @context (and no default given)")
                self.api_version = api_version
            else:
                # determine API version from last context, pick highest
                # API version for a given context by searching highest
                # version first (i.e. get 2.1 not 2.0)
                api_version_read = None
                for v in sorted(CONF.keys(), reverse=True):
                    if (v > '1.0' and self.context == CONF[v]['api_context']):
                        api_version_read = v
                        break
                if api_version_read is None:
                    raise IIIFInfoContextError(
                        "Unknown @context, cannot determine API version (%s)" %
                        (self.context))
                elif api_version is None:
                    self.api_version = api_version_read
                elif api_version != api_version_read:
                    raise IIIFInfoContextError(
                        "Expected API version '%s' but got @context for API version '%s'" %
                        (api_version, api_version_read))
                if self.api_version < '3.0' and len(self.contexts) > 1:
                    raise IIIFInfoContextError("Multiple top-level @contexts not allowed in versions prior to v3.0")
        self.set_version_info()
        #
        # parse remaining JSON top-level keys
        for param in self.params:
            self.read_property(j, param)
        #
        # sanity check for id
        id_key = self.json_key('id')
        if id_key not in j:
            raise IIIFInfoError("Missing %s in info.json" % (id_key))
        return True

    @property
    def formats(self):
        """The pre 3.0 formats property tied to extra_formats."""
        return self.extra_formats

    @formats.setter
    def formats(self, value):
        """Set pre 3.0 formats by writing to alias extra_formats."""
        self.extra_formats = value

    @property
    def qualities(self):
        """The pre 3.0 qualities property tied to extra_qualities."""
        return self.extra_qualities

    @qualities.setter
    def qualities(self, value):
        """Set pre 3.0 qualitiess by writing to alias extra_qualities."""
        self.extra_qualities = value

    @property
    def supports(self):
        """The pre 3.0 supports property tied to extra_features."""
        return self.extra_features

    @supports.setter
    def supports(self, value):
        """Set pre 3.0 supports by writing to alias extra_features."""
        self.extra_features = value



"""IIIF Authentication for the clickthrough interaction pattern.

See <http://iiif.io/api/auth/#clickthrough-interaction-pattern>.
Trivial code to immediately close the login window and
set the account cookie.
"""

from .auth_flask import IIIFAuthFlask


class IIIFAuthClickthrough(IIIFAuthFlask):
    """IIIF Authentication class using a simple clickthrough."""

    def __init__(self, cookie_prefix=None):
        """Initialize IIIFAuthBasic object."""
        super(IIIFAuthClickthrough, self).__init__(cookie_prefix=cookie_prefix)
        self.auth_pattern = 'clickthrough'
        self.header = 'Restricted Material with Terms of Use'
        self.description = 'Use on condition of saying "IIIF is great!"'
        self.confirm_label = 'I agree'

    def login_service_description(self):
        """Clickthrough login service description.

        The login service description _MUST_ include the token service
        description. Additionally, for a clickthroudh loginThe authentication pattern is indicated via the
        profile URI which is built using self.auth_pattern.
        """
        desc = super(IIIFAuthClickthrough, self).login_service_description()
        desc['confirmLabel'] = self.confirm_label
        return desc

    def login_handler(self, config=None, prefix=None, **args):
        """Login handler for clickthrough.

        We simply set the account cooke and close the login
        window, there is not authentication per se.
        """
        return self.set_cookie_close_window_response("clickthrough-ok")

    # No logout for clickthrough
    logout_handler = None



"""Utility functions and classes for Flask server implementation.

Based around IIIFManipulator objects for any manipulations
requested and is thus very slow.

Simeon Warner - 2014--2018
"""

from flask import Flask, request, make_response, redirect, abort, send_file, url_for, send_from_directory

import base64
import configargparse
import json
import logging
import os
import os.path
import re
from string import Template
import sys
try:  # python3
    from urllib.parse import urljoin, quote as urlquote
    from urllib.request import parse_keqv_list, parse_http_list
except ImportError:  # python2
    from urlparse import urljoin
    from urllib import quote as urlquote
    from urllib2 import parse_keqv_list, parse_http_list

from iiif.error import IIIFError
from iiif.request import IIIFRequest, IIIFRequestPathError, IIIFRequestBaseURI
from iiif.info import IIIFInfo


class Config(object):
    """Class to share configuration information in IIIFHandler instances.

    Designed to allow initialization from other Config
    objects and from argparse.Namespace objects in order to create
    independent copies of config data needed for each Flask request
    handler.
    """

    def __init__(self, *args):
        """Initialize new Config object copying all properties from args."""
        for arg in args:
            for k in list(arg.__dict__.keys()):
                self.__dict__[k] = arg.__dict__[k]


def html_page(title="Page Title", body=""):
    """Create HTML page as string."""
    html = "<html>\n<head><title>%s</title></head>\n<body>\n" % (title)
    html += "<h1>%s</h1>\n" % (title)
    html += body
    html += "</body>\n</html>\n"
    return html


def top_level_index_page(config):
    """HTML top-level index page which provides a link to each handler."""
    title = "IIIF Test Server on %s" % (config.host)
    body = "<ul>\n"
    for prefix in sorted(config.prefixes.keys()):
        body += '<li><a href="/%s">%s</a></li>\n' % (prefix, prefix)
    body += "</ul>\n"
    return html_page(title, body)


def identifiers(config):
    """Show list of identifiers for this prefix.

    Handles both the case of local file based identifiers and
    also image generators.

    Arguments:
        config - configuration object in which:
            config.klass_name - 'gen' if a generator function
            config.generator_dir - directory for generator code
            config.image_dir - directory for images

    Returns:
        ids - a list of ids
    """
    ids = []
    if (config.klass_name == 'gen'):
        for generator in os.listdir(config.generator_dir):
            if (generator == '__init__.py'):
                continue
            (gid, ext) = os.path.splitext(generator)
            if (ext == '.py' and
                    os.path.isfile(os.path.join(config.generator_dir, generator))):
                ids.append(gid)
    else:
        for image_file in os.listdir(config.image_dir):
            (iid, ext) = os.path.splitext(image_file)
            if (ext in ['.jpg', '.png', '.tif'] and
                    os.path.isfile(os.path.join(config.image_dir, image_file))):
                ids.append(iid)
    return ids


def prefix_index_page(config):
    """HTML index page for a specific prefix.

    The prefix seen by the client is obtained from config.client_prefix
    as opposed to the local server prefix in config.prefix. Also uses
    the identifiers(config) function to get identifiers available.

    Arguments:
        config - configuration object in which:
            config.client_prefix - URI path prefix seen by client
            config.host - URI host seen by client
            config.api_version - string for api_version
            config.manipulator - string manipulator type
            config.auth_type - string for auth type
            config.include_osd - whether OSD is included
    """
    title = "IIIF Image API services under %s" % (config.client_prefix)
    # details of this prefix handler
    body = '<p>\n'
    body += 'host = %s<br/>\n' % (config.host)
    body += 'api_version = %s<br/>\n' % (config.api_version)
    body += 'manipulator = %s<br/>\n' % (config.klass_name)
    body += 'auth_type = %s\n</p>\n' % (config.auth_type)
    # table of identifiers and example requests
    ids = identifiers(config)
    api_version = config.api_version
    default = 'native' if api_version < '2.0' else 'default'
    body += '<table border="1">\n<tr><th align="left">Image identifier</th>'
    body += '<th> </th><th>full</th>'
    if (config.klass_name != 'dummy'):
        body += '<th>256,256</th>'
        body += '<th>30deg</th>'
        if (config.include_osd):
            body += '<th> </th>'
    body += "</tr>\n"
    max = 'max' if api_version >= '3.0' else 'full'
    for identifier in sorted(ids):
        base = urljoin('/', config.client_prefix + '/' + identifier)
        body += '<tr><th align="left">%s</th>' % (identifier)
        info = base + "/info.json"
        body += '<td><a href="%s">%s</a></td>' % (info, 'info')
        suffix = "full/%s/0/%s" % (max, default)
        body += '<td><a href="%s">%s</a></td>' % (base + '/' + suffix, suffix)
        if (config.klass_name != 'dummy'):
            suffix = "full/256,256/0/%s" % (default)
            body += '<td><a href="%s">%s</a></td>' % (base + '/' + suffix, suffix)
            suffix = "full/100,/30/%s" % (default)
            body += '<td><a href="%s">%s</a></td>' % (base + '/' + suffix, suffix)
            if (config.include_osd):
                body += '<td><a href="%s/osd.html">OSD</a></td>' % (base)
        body += "</tr>\n"
    body += "</table<\n"
    return html_page(title, body)


def host_port_prefix(host, port, prefix):
    """Return URI composed of scheme, server, port, and prefix."""
    uri = "http://" + host
    if (port != 80):
        uri += ':' + str(port)
    if (prefix):
        uri += '/' + prefix
    return uri


# Flask request handlers


def osd_page_handler(config=None, identifier=None, prefix=None, **args):
    """Flask handler to produce HTML response for OpenSeadragon view of identifier.

    Arguments:
        config - Config object for this IIIF handler
        identifier - identifier of image/generator
        prefix - path prefix
        **args - other aguments ignored
    """
    template_dir = os.path.join(os.path.dirname(__file__), 'templates')
    with open(os.path.join(template_dir, 'testserver_osd.html'), 'r') as f:
        template = f.read()
    d = dict(prefix=prefix,
             identifier=identifier,
             api_version=config.api_version,
             osd_version='2.0.0',
             osd_uri='/openseadragon200/openseadragon.min.js',
             osd_images_prefix='/openseadragon200/images',
             osd_height=500,
             osd_width=500,
             info_json_uri='info.json')
    return make_response(Template(template).safe_substitute(d))


class IIIFHandler(object):
    """IIIFHandler class."""

    def __init__(self, prefix, identifier, config, klass, auth):
        """Initialize IIIFHandler setting key configurations.

        Positional parameters:
        prefix -- URI prefix (without leading or trailing slashes)
        identifier -- identifier of image
        config -- instance of Config class
        klass -- IIIFManipulator sub-class to do manipulations
        auth -- IIIFAuth sub-class instance for auth or None
        """
        self.prefix = prefix
        self.identifier = identifier
        self.config = config
        self.klass = klass
        self.api_version = config.api_version
        self.auth = auth
        self.degraded = False
        self.logger = logging.getLogger('IIIFHandler')
        #
        # Create objects to process request
        self.iiif = IIIFRequest(api_version=self.api_version,
                                identifier=self.identifier)
        self.manipulator = klass(api_version=self.api_version)
        #
        # Set up auth object with locations if not already done
        if (self.auth and not self.auth.login_uri):
            self.auth.login_uri = self.server_and_prefix + '/login'
            if (self.auth.logout_handler is not None):
                self.auth.logout_uri = self.server_and_prefix + '/logout'
            self.auth.access_token_uri = self.server_and_prefix + '/token'
        #
        # Response headers
        # -- All responses should have CORS header
        self.headers = {'Access-control-allow-origin': '*'}

    @property
    def server_and_prefix(self):
        """Server and prefix from config."""
        return(host_port_prefix(self.config.host, self.config.port, self.prefix))

    @property
    def json_mime_type(self):
        """Return the MIME type for a JSON response.

        For version 2.0+ the server must return json-ld MIME type if that
        format is requested. Implement for 1.1 also.
        http://iiif.io/api/image/2.1/#information-request
        """
        mime_type = "application/json"
        if (self.api_version >= '1.1' and 'Accept' in request.headers):
            mime_type = do_conneg(request.headers['Accept'], [
                                  'application/ld+json']) or mime_type
        return mime_type

    @property
    def file(self):
        """Filename property for the source image for the current identifier."""
        file = None
        if (self.config.klass_name == 'gen'):
            for ext in ['.py']:
                file = os.path.join(
                    self.config.generator_dir, self.identifier + ext)
                if (os.path.isfile(file)):
                    return file
        else:
            for ext in ['.jpg', '.png', '.tif']:
                file = os.path.join(self.config.image_dir,
                                    self.identifier + ext)
                if (os.path.isfile(file)):
                    return file
        # failed, show list of available identifiers as error
        available = "\n ".join(identifiers(self.config))
        raise IIIFError(code=404, parameter="identifier",
                        text="Image resource '" + self.identifier + "' not found. Local resources available:" + available + "\n")

    def add_compliance_header(self):
        """Add IIIF Compliance level header to response."""
        if (self.manipulator.compliance_uri is not None):
            self.headers['Link'] = '<' + \
                self.manipulator.compliance_uri + '>;rel="profile"'

    def make_response(self, content, code=200, headers=None):
        """Wrapper around Flask.make_response which also adds any local headers."""
        if headers:
            for header in headers:
                self.headers[header] = headers[header]
        return make_response(content, code, self.headers)

    def image_information_response(self):
        """Parse image information request and create response."""
        dr = degraded_request(self.identifier)
        if (dr):
            self.logger.info("image_information: degraded %s -> %s" %
                             (self.identifier, dr))
            self.degraded = self.identifier
            self.identifier = dr
        else:
            self.logger.info("image_information: %s" % (self.identifier))
        # get size
        self.manipulator.srcfile = self.file
        self.manipulator.do_first()
        # most of info.json comes from config, a few things specific to image
        info = {'tile_height': self.config.tile_height,
                'tile_width': self.config.tile_width,
                'scale_factors': self.config.scale_factors
                }
        # calculate scale factors if not hard-coded
        if ('auto' in self.config.scale_factors):
            info['scale_factors'] = self.manipulator.scale_factors(
                self.config.tile_width, self.config.tile_height)
        i = IIIFInfo(conf=info, api_version=self.api_version)
        i.server_and_prefix = self.server_and_prefix
        i.identifier = self.iiif.identifier
        i.width = self.manipulator.width
        i.height = self.manipulator.height
        if (self.api_version >= '2.0'):
            # FIXME - should come from manipulator
            i.qualities = ["default", "color", "gray"]
        else:
            # FIXME - should come from manipulator
            i.qualities = ["native", "color", "gray"]
        i.formats = ["jpg", "png"]  # FIXME - should come from manipulator
        if (self.auth):
            self.auth.add_services(i)

        return self.make_response(i.as_json(),
                                  headers={"Content-Type": self.json_mime_type})

    def image_request_response(self, path):
        """Parse image request and create response."""
        # Parse the request in path
        if (len(path) > 1024):
            raise IIIFError(code=414,
                            text="URI Too Long: Max 1024 chars, got %d\n" % len(path))
        try:
            self.iiif.identifier = self.identifier
            self.iiif.parse_url(path)
        except IIIFRequestPathError as e:
            # Reraise as IIIFError with code=404 because we can't tell
            # whether there was an encoded slash in the identifier or
            # whether there was a bad number of path segments.
            raise IIIFError(code=404, text=e.text)
        except IIIFError as e:
            # Pass through
            raise e
        except Exception as e:
            # Something completely unexpected => 500
            raise IIIFError(code=500,
                            text="Internal Server Error: unexpected exception parsing request (" + str(e) + ")")
        dr = degraded_request(self.identifier)
        if (dr):
            self.logger.info("image_request: degraded %s -> %s" %
                             (self.identifier, dr))
            self.degraded = self.identifier
            self.identifier = dr
            self.iiif.quality = 'gray'
        else:
            # Parsed request OK, attempt to fulfill
            self.logger.info("image_request: %s" % (self.identifier))
        file = self.file
        self.manipulator.srcfile = file
        self.manipulator.do_first()
        if (self.api_version < '2.0' and
                self.iiif.format is None and
                'Accept' in request.headers):
            # In 1.0 and 1.1 conneg was specified as an alternative to format, see:
            # http://iiif.io/api/image/1.0/#format
            # http://iiif.io/api/image/1.1/#parameters-format
            formats = {'image/jpeg': 'jpg', 'image/tiff': 'tif',
                       'image/png': 'png', 'image/gif': 'gif',
                       'image/jp2': 'jps', 'application/pdf': 'pdf'}
            accept = do_conneg(request.headers['Accept'], list(formats.keys()))
            # Ignore Accept header if not recognized, should this be an error
            # instead?
            if (accept in formats):
                self.iiif.format = formats[accept]
        (outfile, mime_type) = self.manipulator.derive(file, self.iiif)
        # FIXME - find efficient way to serve file with headers
        # could this be the answer: https://stackoverflow.com/questions/31554680/how-to-send-header-in-flask-send-file
        # currently no headers are sent with the file
        self.add_compliance_header()
        return self.make_response(send_file(outfile, mimetype=mime_type))

    def error_response(self, e):
        """Make response for an IIIFError e.

        Also add compliance header.
        """
        self.add_compliance_header()
        return self.make_response(*e.image_server_response(self.api_version))


def iiif_info_handler(prefix=None, identifier=None,
                      config=None, klass=None, auth=None, **args):
    """Handler for IIIF Image Information requests."""
    if (not auth or degraded_request(identifier) or auth.info_authz()):
        # go ahead with request as made
        if (auth):
            logging.debug("Authorized for image %s" % identifier)
        i = IIIFHandler(prefix, identifier, config, klass, auth)
        try:
            return i.image_information_response()
        except IIIFError as e:
            return i.error_response(e)
    elif (auth.info_authn()):
        # authn but not authz -> 401
        abort(401)
    else:
        # redirect to degraded
        response = redirect(host_port_prefix(
            config.host, config.port, prefix) + '/' + identifier + '-deg/info.json')
        response.headers['Access-control-allow-origin'] = '*'
        return response
iiif_info_handler.provide_automatic_options = False


def iiif_image_handler(prefix=None, identifier=None,
                       path=None, config=None, klass=None, auth=None, **args):
    """Handler for IIIF Image Requests.

    Behaviour for case of a non-authn or non-authz case is to
    return 403.
    """
    if (not auth or degraded_request(identifier) or auth.image_authz()):
        # serve image
        if (auth):
            logging.debug("Authorized for image %s" % identifier)
        i = IIIFHandler(prefix, identifier, config, klass, auth)
        try:
            return i.image_request_response(path)
        except IIIFError as e:
            return i.error_response(e)
    else:
        # redirect to degraded (for not authz and for authn but not authz too)
        degraded_uri = host_port_prefix(
            config.host, config.port, prefix) + '/' + identifier + '-deg/' + path
        logging.info("Redirection to degraded: %s" % degraded_uri)
        response = redirect(degraded_uri)
        response.headers['Access-control-allow-origin'] = '*'
        return response
iiif_image_handler.provide_automatic_options = False


def degraded_request(identifier):
    """Return True (non-degraded id) if this is a degraded request, False otherwise."""
    if identifier.endswith('-deg'):
        return identifier[:-4]
    return False


def options_handler(**args):
    """Handler to respond to OPTIONS preflight CORS requests."""
    headers = {'Access-Control-Allow-Origin': '*',
               'Access-Control-Allow-Methods': 'GET,OPTIONS',
               'Access-Control-Allow-Headers': 'Origin, Accept, Accept-Encoding, Authorization'}
    return make_response("", 200, headers)


def parse_accept_header(accept):
    """Parse an HTTP Accept header.

    Parses *accept*, returning a list with pairs of
    (media_type, q_value), ordered by q values.

    Adapted from <https://djangosnippets.org/snippets/1042/>
    """
    result = []
    for media_range in accept.split(","):
        parts = media_range.split(";")
        media_type = parts.pop(0).strip()
        media_params = []
        q = 1.0
        for part in parts:
            (key, value) = part.lstrip().split("=", 1)
            if key == "q":
                q = float(value)
            else:
                media_params.append((key, value))
        result.append((media_type, tuple(media_params), q))
    result.sort(key=lambda x: -x[2])
    return result


def parse_authorization_header(value):
    """Parse the Authenticate header.

    Returns nothing on failure, opts hash on success with type='basic' or 'digest'
    and other params.

    <http://nullege.com/codes/search/werkzeug.http.parse_authorization_header>
    <http://stackoverflow.com/questions/1349367/parse-an-http-request-authorization-header-with-python>
    <http://bugs.python.org/file34041/0001-Add-an-authorization-header-to-the-initial-request.patch>
    """
    try:
        (auth_type, auth_info) = value.split(' ', 1)
        auth_type = auth_type.lower()
    except ValueError:
        return
    if (auth_type == 'basic'):
        try:
            decoded = base64.b64decode(auth_info).decode(
                'utf-8')  # b64decode gives bytes in python3
            (username, password) = decoded.split(':', 1)
        except (ValueError, TypeError):  # py3, py2
            return
        return {'type': 'basic', 'username': username, 'password': password}
    elif (auth_type == 'digest'):
        try:
            auth_map = parse_keqv_list(parse_http_list(auth_info))
        except ValueError:
            return
        logging.debug(auth_map)
        for key in 'username', 'realm', 'nonce', 'uri', 'response':
            if key not in auth_map:
                return
        if 'qop' in auth_map and ('nc' not in auth_map or 'cnonce' not in auth_map):
            return
        auth_map['type'] = 'digest'
        return auth_map
    else:
        # unknown auth type
        return


def do_conneg(accept, supported):
    """Parse accept header and look for preferred type in supported list.

    Arguments:
        accept - HTTP Accept header
        supported - list of MIME type supported by the server

    Returns:
        supported MIME type with highest q value in request, else None.

    FIXME - Should replace this with negotiator2
    """
    for result in parse_accept_header(accept):
        mime_type = result[0]
        if (mime_type in supported):
            return mime_type
    return None

######################################################################


def setup_auth_paths(app, auth, prefix, params):
    """Add URL rules for auth paths."""
    base = urljoin('/', prefix + '/')  # Must end in slash
    app.add_url_rule(base + 'login', prefix + 'login_handler',
                     auth.login_handler, defaults=params)
    app.add_url_rule(base + 'logout', prefix + 'logout_handler',
                     auth.logout_handler, defaults=params)
    if (auth.client_id_handler):
        app.add_url_rule(base + 'client', prefix + 'client_id_handler',
                         auth.client_id_handler, defaults=params)
    app.add_url_rule(base + 'token', prefix + 'access_token_handler',
                     auth.access_token_handler, defaults=params)
    if (auth.home_handler):
        app.add_url_rule(base + 'home', prefix + 'home_handler',
                         auth.home_handler, defaults=params)


def make_prefix(api_version, manipulator, auth_type):
    """Make prefix string based on configuration parameters."""
    prefix = "%s_%s" % (api_version, manipulator)
    if (auth_type and auth_type != 'none'):
        prefix += '_' + auth_type
    return prefix


def split_comma_argument(comma_sep_str):
    """Split a comma separated option into a list."""
    terms = []
    for term in comma_sep_str.split(','):
        if term:
            terms.append(term)
    return terms


def add_shared_configs(p, base_dir=''):
    """Add configargparser/argparse configs for shared argument.

    Arguments:
        p - configargparse.ArgParser object
        base_dir - base directory for file/path defaults.
    """
    p.add('--host', default='localhost',
          help="Service host")
    p.add('--port', '-p', type=int, default=8000,
          help="Service port")
    p.add('--app-host', default=None,
          help="Local application host for reverse proxy deployment, "
               "as opposed to service --host (must also specify --app-port)")
    p.add('--app-port', type=int, default=None,
          help="Local application port for reverse proxy deployment, "
               "as opposed to service --port (must also specify --app-host)")
    p.add('--image-dir', '-d', default=os.path.join(base_dir, 'testimages'),
          help="Image file directory")
    p.add('--generator-dir', default=os.path.join(base_dir, 'iiif/generators'),
          help="Generator directory for manipulator='gen'")
    p.add('--tile-height', type=int, default=512,
          help="Tile height")
    p.add('--tile-width', type=int, default=512,
          help="Tile width")
    p.add('--gauth-client-secret', default=os.path.join(base_dir, 'client_secret.json'),
          help="Name of file with Google auth client secret")
    p.add('--include-osd', action='store_true',
          help="Include a page with OpenSeadragon for each source")
    p.add('--access-cookie-lifetime', type=int, default=3600,
          help="Set access cookie lifetime for authenticated access in seconds")
    p.add('--access-token-lifetime', type=int, default=10,
          help="Set access token lifetime for authenticated access in seconds")
    p.add('--config', is_config_file=True, default=None,
          help='Read config from given file path')
    p.add('--debug', action='store_true',
          help="Set debug mode for web application. INSECURE!")
    p.add('--verbose', '-v', action='store_true',
          help="Be verbose")
    p.add('--quiet', '-q', action='store_true',
          help="Minimal output only")


def add_handler(app, config):
    """Add a single handler to the app.

    Adds one IIIF Image API handler to app, with config from config.

    Arguments:
        app - Flask app
        config - Configuration object in which:
            config.prefix - String path prefix for this handler
            config.client_prefix - String path prefix seen by client (which may be different
                because of reverse proxy or such
            config.klass_name - Manipulator class, e.g. 'pil'
            config.api_version - e.g. '2.1'
            config.include_osd - True or False to include OSD
            config.gauth_client_secret_file - filename if auth_type='gauth'
            config.access_cookie_lifetime - number of seconds
            config.access_token_lifetime - number of seconds
            config.auth_type - Auth type string or 'none'

    Returns True on success, nothing otherwise.
    """
    auth = None
    if (config.auth_type is None or config.auth_type == 'none'):
        pass
    elif (config.auth_type == 'gauth'):
        from iiif.auth_google import IIIFAuthGoogle
        auth = IIIFAuthGoogle(client_secret_file=config.gauth_client_secret_file)
    elif (config.auth_type == 'basic'):
        from iiif.auth_basic import IIIFAuthBasic
        auth = IIIFAuthBasic()
    elif (config.auth_type == 'clickthrough'):
        from iiif.auth_clickthrough import IIIFAuthClickthrough
        auth = IIIFAuthClickthrough()
    elif (config.auth_type == 'kiosk'):
        from iiif.auth_kiosk import IIIFAuthKiosk
        auth = IIIFAuthKiosk()
    elif (config.auth_type == 'external'):
        from iiif.auth_external import IIIFAuthExternal
        auth = IIIFAuthExternal()
    else:
        logging.error("Unknown auth type %s, ignoring" % (config.auth_type))
        return
    if (auth is not None):
        auth.access_cookie_lifetime = config.access_cookie_lifetime
        auth.access_token_lifetime = config.access_token_lifetime
    klass = None
    if (config.klass_name == 'pil'):
        from iiif.manipulator_pil import IIIFManipulatorPIL
        klass = IIIFManipulatorPIL
    elif (config.klass_name == 'netpbm'):
        from iiif.manipulator_netpbm import IIIFManipulatorNetpbm
        klass = IIIFManipulatorNetpbm
    elif (config.klass_name == 'dummy'):
        from iiif.manipulator import IIIFManipulator
        klass = IIIFManipulator
    elif (config.klass_name == 'gen'):
        from iiif.manipulator_gen import IIIFManipulatorGen
        klass = IIIFManipulatorGen
    else:
        logging.error("Unknown manipulator type %s, ignoring" % (config.klass_name))
        return
    base = urljoin('/', config.prefix + '/')  # ensure has trailing slash
    client_base = urljoin('/', config.client_prefix + '/')  # ensure has trailing slash
    logging.warning("Installing %s IIIFManipulator at %s v%s %s" %
                    (config.klass_name, base, config.api_version, config.auth_type))
    params = dict(config=config, klass=klass, auth=auth, prefix=config.client_prefix)
    app.add_url_rule(base.rstrip('/'), 'prefix_index_page',
                     prefix_index_page, defaults={'config': config})
    app.add_url_rule(base, 'prefix_index_page',
                     prefix_index_page, defaults={'config': config})
    app.add_url_rule(base + '<string(minlength=1):identifier>/info.json',
                     'options_handler', options_handler, methods=['OPTIONS'])
    app.add_url_rule(base + '<string(minlength=1):identifier>/info.json',
                     'iiif_info_handler', iiif_info_handler, methods=['GET'], defaults=params)
    if (config.include_osd):
        app.add_url_rule(base + '<string(minlength=1):identifier>/osd.html',
                         'osd_page_handler', osd_page_handler, methods=['GET'], defaults=params)
    app.add_url_rule(base + '<string(minlength=1):identifier>/<path:path>',
                     'iiif_image_handler', iiif_image_handler, methods=['GET'], defaults=params)
    if (auth):
        setup_auth_paths(app, auth, config.client_prefix, params)
    # redirects to info.json must come after auth
    app.add_url_rule(base + '<string(minlength=1):identifier>',
                     'iiif_info_handler',
                     redirect_to=client_base + '<identifier>/info.json')
    app.add_url_rule(base + '<string(minlength=1):identifier>/',
                     'iiif_info_handler',
                     redirect_to=client_base + '<identifier>/info.json')
    return True


def serve_static(filename=None, prefix='', basedir=None):
    """Handler for static files: server filename in basedir/prefix.

    If not specified then basedir defaults to the local third_party
    directory.
    """
    if not basedir:
        basedir = os.path.join(os.path.dirname(__file__), 'third_party')
    return send_from_directory(os.path.join(basedir, prefix), filename)


class ReverseProxied(object):
    """Wrap the application call to deal with a reverse proxy setup.

    Overrides HTTP_HOST environment setting.
    See: <http://flask.pocoo.org/snippets/35/>
    """

    def __init__(self, app, host):
        """Initialize reverse proxy wrapper, store host.

        Arguments:
            app - the application being reverse proxied
            host - the configured host name for the application
        """
        self.app = app
        self.host = host

    def __call__(self, environ, start_response):
        """Set environment with service host."""
        environ['HTTP_HOST'] = self.host
        return self.app(environ, start_response)


def write_pid_file():
    """Write a file with the PID of this server instance.

    Call when setting up a command line testserver.
    """
    pidfile = os.path.basename(sys.argv[0])[:-3] + '.pid'  # strip .py, add .pid
    with open(pidfile, 'w') as fh:
        fh.write("%d\n" % os.getpid())
        fh.close()


def setup_app(app, cfg):
    """Setup Flask app and handle reverse proxy setup if configured.

    Arguments:
        app - Flask application
        cfg - configuration data
    """
    # Set up app_host and app_port in case that we are running
    # under reverse proxy setup, otherwise they default to
    # config.host and config.port.
    if (cfg.app_host and cfg.app_port):
        logging.warning("Reverse proxy for service at http://%s:%d/ ..." % (cfg.host, cfg.port))
        app.wsgi_app = ReverseProxied(app.wsgi_app, cfg.host)
    elif (cfg.app_host or cfg.app_port):
        logging.critical("Must specify both app-host and app-port for reverse proxy configuration, aborting")
        sys.exit(1)
    else:
        cfg.app_host = cfg.host
        cfg.app_port = cfg.port
    logging.warning("Setup server on http://%s:%d/ ..." % (cfg.app_host, cfg.app_port))
    return(app)



#!/usr/bin/env python
"""Crude webserver that services IIIF Image API requests.

Relies upon IIIFManipulator objects to do any manipulations
requested and is thus very slow. Supports a number of different
versions of the specification via different base URIs (prefixes).

Simeon Warner - 2014--2018
"""

from flask import Flask
import logging
import re
import configargparse
import os
import os.path
import sys

from iiif.flask_utils import Config, write_pid_file, add_handler, make_prefix, top_level_index_page, serve_static, ReverseProxied, setup_app, split_comma_argument, add_shared_configs


def get_config(base_dir=''):
    """Get config from defaults, config file and/or parse arguments.

    Uses configargparse to allow argments to be set from a config file
    or via command line arguments.

      base_dir - set a specific base directory for file/path defaults.
    """
    p = configargparse.ArgParser(description='IIIF Image Testserver',
                                 default_config_files=[os.path.join(base_dir, 'iiif_testserver.cfg')],
                                 formatter_class=configargparse.ArgumentDefaultsHelpFormatter)
    add_shared_configs(p, base_dir)
    p.add('--container-prefix', default='',
          help="Container prefix seen by client to add to links generated")
    p.add('--scale-factors', default='auto',
          help="Set of tile scale factors or 'auto' to calculate for each image "
               "such that there are tiles up to the full image")
    p.add('--api-versions', default='1.0,1.1,2.0,2.1,3.0',
          help="Set of API versions to support")
    p.add('--manipulators', default='pil',
          help="Set of manipuators to instantiate. May be dummy,netpbm,pil "
               "or gen for generated image")
    p.add('--auth-types', default='none',
          help="Set of authentication types to support")
    p.add('--pages-dir', default=os.path.join(base_dir, 'testpages'),
          help="Test pages directory")
    p.add('--auth', action='store_true',
          help="Enable features implementing the IIIF Authentication specification")
    args = p.parse_args()

    if (args.debug):
        args.verbose = True
    elif (args.verbose):
        args.quiet = False

    # Split list arguments
    args.scale_factors = split_comma_argument(args.scale_factors)
    args.manipulators = split_comma_argument(args.manipulators)
    args.api_versions = split_comma_argument(args.api_versions)
    args.auth_types = split_comma_argument(args.auth_types)

    # Authentication features...
    if (args.auth and 'gauth' not in args.auth_types):
        args.auth_types.append('gauth')
    if (args.auth and 'basic' not in args.auth_types):
        args.auth_types.append('basic')
    if (args.auth and 'clickthrough' not in args.auth_types):
        args.auth_types.append('clickthrough')
    if (args.auth and 'kiosk' not in args.auth_types):
        args.auth_types.append('kiosk')
    if (args.auth and 'external' not in args.auth_types):
        args.auth_types.append('external')

    logging_level = logging.WARNING
    if args.verbose:
        logging_level = logging.INFO
    elif args.quiet:
        logging_level = logging.ERROR
    logging.basicConfig(format='%(name)s: %(message)s', level=logging_level)

    return(args)


def create_testserver_flask_app(cfg):
    """Create testserver Flask application with one or more IIIF handlers."""
    # Create Flask app
    app = Flask(__name__, static_url_path='/' + cfg.pages_dir)
    Flask.secret_key = "SECRET_HERE"
    app.debug = cfg.debug

    # Create shared configuration dict based on options
    config = Config(cfg)
    config.homedir = os.path.dirname(os.path.realpath(__file__))
    config.gauth_client_secret_file = os.path.join(
        config.homedir, config.gauth_client_secret)

    # Install request handlers
    client_prefixes = dict()
    for api_version in cfg.api_versions:
        for klass_name in cfg.manipulators:
            for auth_type in cfg.auth_types:
                # auth only for >=2.1
                if (auth_type != 'none' and float(api_version) < 2.1):
                    continue
                handler_config = Config(config)
                handler_config.api_version = api_version
                handler_config.klass_name = klass_name
                handler_config.auth_type = auth_type
                prefix = make_prefix(api_version, klass_name, auth_type)
                client_prefix = os.path.join(config.container_prefix, prefix)
                logging.debug("prefix = %s, client_prefix = %s" % (prefix, client_prefix))
                client_prefixes[client_prefix] = prefix
                handler_config.prefix = prefix
                handler_config.client_prefix = client_prefix
                add_handler(app, handler_config)

    # Index page
    config.prefixes = client_prefixes
    app.add_url_rule('/', 'top_level_index_page',
                     top_level_index_page, defaults={'config': config})

    if cfg.include_osd:
        # OpenSeadragon files
        app.add_url_rule('/openseadragon200/<path:filename>', 'OSD pages', serve_static,
                         defaults={'prefix': 'openseadragon200'})

    return(app)


if __name__ == '__main__':
    # Command line, run server
    write_pid_file()
    cfg = get_config()
    app = setup_app(create_testserver_flask_app(cfg), cfg)
    app.run(host=cfg.app_host, port=cfg.app_port)



"""IIIF Authentication for the kiosk interaction pattern.

See <http://iiif.io/api/auth/#kiosk-interaction-pattern>.
Trivial code to immediately close the login window and
set the account cookie.
"""

from .auth_flask import IIIFAuthFlask


class IIIFAuthKiosk(IIIFAuthFlask):
    """IIIF Authentication class with no user interaction."""

    def __init__(self, cookie_prefix=None):
        """Initialize IIIFAuthBasic object."""
        super(IIIFAuthKiosk, self).__init__(cookie_prefix=cookie_prefix)
        self.auth_pattern = 'kiosk'
        self.header = 'Kiosk setup in progress, please wait...'

    def login_handler(self, config=None, prefix=None, **args):
        """Login handler for kiosk.

        Override this method to do any checks to do kiosk auth, e.g. we might
        test the orginator of the request or similar. In this null
        implementation we simply set the account cookie and close the login
        window, there is no authentication per se.
        """
        return self.set_cookie_close_window_response("kiosk-null-ok")

    # No logout for kiosk pattern
    logout_handler = None



"""Support for IIIF Authentication."""

import json
import logging
import random
import re
import string
import time


class IIIFAuth(object):
    """IIIF Authentication Class.

    Base authentication class which does not actually implement an
    authentication scheme but is a foundation for specific sub-classes
    for different schemes.
    """

    def __init__(self, cookie_prefix=None):
        """Initialize IIIFAuth object."""
        self.profile_base = 'http://iiif.io/api/auth/1/'
        self.name = 'image server'
        self.auth_pattern = 'login'
        self.header = None
        self.description = None
        self.auth_type = None
        self.login_uri = None
        self.logout_uri = None
        self.client_id_uri = None
        self.access_token_uri = None
        self.logger = logging.getLogger(__name__)
        # Need to have different cookie names for each auth domain
        # running on the same server
        self.set_cookie_prefix(cookie_prefix)
        self.access_cookie_name = self.cookie_prefix + 'access'
        self.access_cookies = {}
        self.access_cookie_lifetime = 36000  # seconds
        self.access_tokens = {}
        self.access_token_lifetime = 600  # seconds

    def set_cookie_prefix(self, cookie_prefix=None):
        """Set a random cookie prefix unless one is specified.

        In order to run multiple demonstration auth services on the
        same server we need to have different cookie names for each
        auth domain. Unless cookie_prefix is set, generate a random
        one.
        """
        if (cookie_prefix is None):
            self.cookie_prefix = "%06d_" % int(random.random() * 1000000)
        else:
            self.cookie_prefix = cookie_prefix

    def add_services(self, info):
        """Add auth service descriptions to an IIIFInfo object.

        Login service description is the wrapper for all other
        auth service descriptions so we have nothing unless
        self.login_uri is specified. If we do then add any other
        auth services at children.

        Exactly the same structure is used in the authorized
        and unauthorized cases (although in the data could be
        different).
        """
        if (self.login_uri):
            svc = self.login_service_description()
            svcs = []
            if (self.logout_uri):
                svcs.append(self.logout_service_description())
            if (self.client_id_uri):
                svcs.append(self.client_id_service_description())
            if (self.access_token_uri):
                svcs.append(self.access_token_service_description())
            # Add one as direct child of service property, else array for >1
            if (len(svcs) == 1):
                svc['service'] = svcs[0]
            elif (len(svcs) > 1):
                svc['service'] = svcs
            info.add_service(svc)

    def login_service_description(self):
        """Login service description.

        The login service description _MUST_ include the token service
        description. The authentication pattern is indicated via the
        profile URI which is built using self.auth_pattern.
        """
        label = 'Login to ' + self.name
        if (self.auth_type):
            label = label + ' (' + self.auth_type + ')'
        desc = {"@id": self.login_uri,
                "profile": self.profile_base + self.auth_pattern,
                "label": label}
        if (self.header):
            desc['header'] = self.header
        if (self.description):
            desc['description'] = self.description
        return desc

    def logout_service_description(self):
        """Logout service description."""
        label = 'Logout from ' + self.name
        if (self.auth_type):
            label = label + ' (' + self.auth_type + ')'
        return({"@id": self.logout_uri,
                "profile": self.profile_base + 'logout',
                "label": label})

    def client_id_service_description(self):
        """Client Id service description."""
        return({"@id": self.client_id_uri,
                "profile": self.profile_base + 'clientId'})

    def access_token_service_description(self):
        """Access Token service description."""
        return({"@id": self.access_token_uri,
                "profile": self.profile_base + 'token'})

    # Override with method to implement
    access_token_handler = None

    def access_token_response(self, token, message_id=None):
        """Access token response structure.

        Success if token is set, otherwise (None, empty string) give error
        response. If message_id is set then an extra messageId attribute is
        set in the response to handle postMessage() responses.
        """
        if (token):
            data = {"accessToken": token,
                    "expiresIn": self.access_token_lifetime}
            if (message_id):
                data['messageId'] = message_id
        else:
            data = {"error": "client_unauthorized",
                    "description": "No authorization details received"}
        return data

    # Override with method to implement
    client_id_handler = None

    # Override with method to implement
    home_handler = None

    def info_authn(self):
        """Check to see if user is authenticated for info.json.

        Null implementation that always returns False, must override
        to implement authorization.
        """
        return False

    def info_authz(self):
        """Check to see if user is authenticated and authorized for info.json.

        Null implementation that says that any authenticated user is
        authorized.
        """
        return self.info_authn()

    def image_authn(self):
        """Check to see if user is authenticated for image requests.

        Null implementation that always returns False, must override
        to implement authorization.
        """
        return False

    def image_authz(self):
        """Check to see if user is authenticated and authorized for image requests.

        Null implementation that says that any authenticated user is
        authorized.
        """
        return self.image_authn()

    def scheme_host_port_prefix(self, scheme='http', host='host',
                                port=None, prefix=None):
        """Return URI composed of scheme, server, port, and prefix."""
        uri = scheme + '://' + host
        if (port and not ((scheme == 'http' and port == 80) or
                          (scheme == 'https' and port == 443))):
            uri += ':' + str(port)
        if (prefix):
            uri += '/' + prefix
        return uri

    def _generate_random_string(self, container, length=20):
        """Generate a random cookie or token string not in container.

        The cookie or token should be secure in the sense that it should not
        be likely to be able guess a value. Because it is not derived from
        anything else, there is no vulnerability of the token from computation,
        or possible leakage of information from the token.
        """
        while True:
            s = ''.join([random.SystemRandom().choice(string.digits + string.ascii_letters)
                         for n in range(length)])
            if (s not in container):
                break
        return s

    def account_allowed(self, account):
        """True if the account credentials should be accepted.

        Default implementation is that any account is allowed,
        so response is True if account is True. Override this method
        to authorize particular values of account.
        """
        return True if (account) else False

    def access_cookie(self, account):
        """Make and store access cookie for a given account.

        If account is allowed then make a cookie and add it to the dict
        of accepted access cookies with current timestamp as the value.
        Return the access cookie.

        Otherwise return None.
        """
        if (self.account_allowed(account)):
            cookie = self._generate_random_string(self.access_cookies)
            self.access_cookies[cookie] = int(time.time())
            return cookie
        else:
            return None

    def access_cookie_valid(self, cookie, log_msg):
        """Check access cookie validity.

        Returns true if the access cookie is valid. The set of allowed
        access cookies is stored in self.access_cookies.

        Uses log_msg as prefix to info level log message of accetance or
        rejection.
        """
        if (cookie in self.access_cookies):
            age = int(time.time()) - self.access_cookies[cookie]
            if (age <= (self.access_cookie_lifetime + 1)):
                self.logger.info(log_msg + " " + cookie +
                                 " ACCEPTED COOKIE (%ds old)" % age)
                return True
            # Expired...
            self.logger.info(log_msg + " " + cookie +
                             " EXPIRED COOKIE (%ds old > %ds)" %
                             (age, self.access_cookie_lifetime))
            # Keep cookie for 2x lifetim in order to generate
            # helpful expired message
            if (age > (self.access_cookie_lifetime * 2)):
                del self.access_cookies[cookie]
            return False
        else:
            self.logger.info(log_msg + " " + cookie + " REJECTED COOKIE")
            return False

    def access_token(self, cookie):
        """Make and store access token as proxy for the access cookie.

        Create an access token to act as a proxy for access cookie, add it to
        the dict of accepted access tokens with (cookie, current timestamp)
        as the value. Return the access token. Return None if cookie is not set.
        """
        if (cookie):
            token = self._generate_random_string(self.access_tokens)
            self.access_tokens[token] = (cookie, int(time.time()))
            return token
        else:
            return None

    def access_token_valid(self, token, log_msg):
        """Check token validity.

        Returns true if the token is valid. The set of allowed access tokens
        is stored in self.access_tokens.

        Uses log_msg as prefix to info level log message of acceptance or
        rejection.
        """
        if (token in self.access_tokens):
            (cookie, issue_time) = self.access_tokens[token]
            age = int(time.time()) - issue_time
            if (age <= (self.access_token_lifetime + 1)):
                self.logger.info(log_msg + " " + token +
                                 " ACCEPTED TOKEN (%ds old)" % age)
                return True
            # Expired...
            self.logger.info(log_msg + " " + token +
                             " EXPIRED TOKEN (%ds old > %ds)" %
                             (age, self.access_token_lifetime))
            # Keep token for 2x lifetim in order to generate
            # helpful expired message
            if (age > (self.access_token_lifetime * 2)):
                del self.access_tokens[token]
            return False
        else:
            self.logger.info(log_msg + " " + token + " REJECTED TOKEN")
            return False



"""IIIF Authentication using HTTP Basic auth.

See <http://iiif.io/api/auth/#login-interaction-pattern>.

This code extends IIIFAuthFlask and is thus specific to the
Flask framework.
"""

import json
import re
import os.path
from flask import request, make_response

from .auth_flask import IIIFAuthFlask


class IIIFAuthBasic(IIIFAuthFlask):
    """IIIF Authentication Class using HTTP Basic Auth."""

    def __init__(self, cookie_prefix=None):
        """Initialize IIIFAuthBasic object."""
        super(IIIFAuthBasic, self).__init__(cookie_prefix=cookie_prefix)
        self.auth_pattern = 'login'
        self.auth_type = 'basic auth'

    def login_handler(self, config=None, prefix=None, **args):
        """HTTP Basic login handler.

        Respond with 401 and WWW-Authenticate header if there are no
        credentials or bad credentials. If there are credentials then
        simply check for username equal to password for validity.
        """
        headers = {}
        headers['Access-control-allow-origin'] = '*'
        headers['Content-type'] = 'text/html'
        auth = request.authorization
        if (auth and auth.username == auth.password):
            return self.set_cookie_close_window_response(
                "valid-http-basic-login")
        else:
            headers['WWW-Authenticate'] = (
                'Basic realm="HTTP-Basic-Auth at %s (u=p to login)"' %
                (self.name))
            return make_response("", 401, headers)



"""IIIF Authentication for a Flask application.

Extends iiif.auth to add Flask web application interaction
necessary to build a real auth system. In theory one
might be able to replace this module as the middle
component to work with another framework.
"""

import json
from flask import request, make_response, redirect

from .auth import IIIFAuth


class IIIFAuthFlask(IIIFAuth):
    """IIIF Authentication Class for a Flask application."""

    def __init__(self, cookie_prefix=None):
        """Initialize IIIFAuthFlask object."""
        super(IIIFAuthFlask, self).__init__(cookie_prefix=cookie_prefix)
        self.account_cookie_name = self.cookie_prefix + 'account'

    def info_authn(self):
        """Check to see if user if authenticated for info.json.

        Must have Authorization header with value that has the form
        "Bearer TOKEN", where TOKEN is an appropriate and valid access
        token.
        """
        authz_header = request.headers.get('Authorization', '[none]')
        if (not authz_header.startswith('Bearer ')):
            return False
        token = authz_header[7:]
        return self.access_token_valid(
            token, "info_authn: Authorization header")

    def image_authn(self):
        """Check to see if user if authenticated for image requests.

        Must have access cookie with an appropriate value.
        """
        authn_cookie = request.cookies.get(
            self.access_cookie_name, default='[none]')
        return self.access_cookie_valid(authn_cookie, "image_authn: auth cookie")

    def login_handler_redirect(self, url):
        """Complete login handler behavior with redirect.

        There will be some authentication specific code before but
        it will return the value from this method which builds the
        Flask redirect response.
        """
        response = redirect(url)
        response.headers['Access-control-allow-origin'] = '*'
        return response

    def logout_handler(self, **args):
        """Handler for logout button.

        Delete cookies and return HTML that immediately closes window
        """
        response = make_response(
            "<html><script>window.close();</script></html>", 200,
            {'Content-Type': "text/html"})
        response.set_cookie(self.account_cookie_name, expires=0)
        response.set_cookie(self.access_cookie_name, expires=0)
        response.headers['Access-Control-Allow-Origin'] = '*'
        return response

    def access_token_handler(self, **args):
        """Get access token based on cookie sent with this request.

        This handler deals with two cases:

        1) Non-browser client (indicated by no messageId set in request)
        where the response is a simple JSON response.

        2) Browser client (indicate by messageId setin request) where
        the request must be made from a an iFrame and the response is
        sent as JSON wrapped in HTML containing a postMessage() script
        that conveys the access token to the viewer.
        """
        message_id = request.args.get('messageId', default=None)
        origin = request.args.get('origin', default='unknown_origin')
        self.logger.info("access_token_handler: origin = " + origin)
        account = request.cookies.get(self.account_cookie_name, default='')
        token = self.access_token(account)

        # Build JSON response
        data_str = json.dumps(self.access_token_response(token, message_id))
        ct = "application/json"

        # If message_id is set the wrap in HTML with postMessage JavaScript
        # for a browser client
        if (message_id is not None):
            data_str = """<html>
<body style="margin: 0px;">
<div>postMessage ACCESS TOKEN %s</div>
<script>
window.parent.postMessage(%s, '%s');
</script>
</body>
</html>
""" % (token, data_str, origin)
            ct = "text/html"

        # Send response along with cookie
        response = make_response(data_str, 200, {'Content-Type': ct})
        if (token):
            self.logger.info(
                "access_token_handler: setting access token = " + token)
            # Set the cookie for the image content
            cookie = self.access_cookie(token)
            self.logger.info(
                "access_token_handler: setting access cookie = " + cookie)
            response.set_cookie(self.access_cookie_name, cookie)
        else:
            self.logger.info(
                "access_token_handler: auth failed, sending error")
        response.headers['Access-control-allow-origin'] = '*'
        return response

    def set_cookie_close_window_response(self, account_cookie_value):
        """Response to set account cookie and close window HTML/JavaScript."""
        response = make_response(
            "<html><script>window.close();</script></html>", 200,
            {'Content-Type': "text/html"})
        response.set_cookie(self.account_cookie_name,
                            account_cookie_value)
        return response

    def request_args_get(self, name, default=''):
        """Wrapper for request.args.get()."""
        return request.args.get(name, default)
